"""Tests for our debugging module."""
from unittest import mock

import pytest

from flake8.main import debug
from flake8.options import manager


def test_dependencies():
    """Verify that we format our dependencies appropriately."""
    assert [] == debug.dependencies()


@pytest.mark.parametrize(
    "plugins, expected",
    [
        ([], []),
        (
            [manager.PluginVersion("pycodestyle", "2.0.0", False)],
            [
                {
                    "plugin": "pycodestyle",
                    "version": "2.0.0",
                    "is_local": False,
                }
            ],
        ),
        (
            [
                manager.PluginVersion("pycodestyle", "2.0.0", False),
                manager.PluginVersion("mccabe", "0.5.9", False),
            ],
            [
                {"plugin": "mccabe", "version": "0.5.9", "is_local": False},
                {
                    "plugin": "pycodestyle",
                    "version": "2.0.0",
                    "is_local": False,
                },
            ],
        ),
        (
            [
                manager.PluginVersion("pycodestyle", "2.0.0", False),
                manager.PluginVersion("my-local", "0.0.1", True),
                manager.PluginVersion("mccabe", "0.5.9", False),
            ],
            [
                {"plugin": "mccabe", "version": "0.5.9", "is_local": False},
                {"plugin": "my-local", "version": "0.0.1", "is_local": True},
                {
                    "plugin": "pycodestyle",
                    "version": "2.0.0",
                    "is_local": False,
                },
            ],
        ),
    ],
)
def test_plugins_from(plugins, expected):
    """Test that we format plugins appropriately."""
    option_manager = mock.Mock(registered_plugins=set(plugins))
    assert expected == debug.plugins_from(option_manager)


@mock.patch("platform.python_implementation", return_value="CPython")
@mock.patch("platform.python_version", return_value="3.5.3")
@mock.patch("platform.system", return_value="Linux")
def test_information(system, pyversion, pyimpl):
    """Verify that we return all the information we care about."""
    expected = {
        "version": "3.1.0",
        "plugins": [
            {"plugin": "mccabe", "version": "0.5.9", "is_local": False},
            {"plugin": "pycodestyle", "version": "2.0.0", "is_local": False},
        ],
        "dependencies": [],
        "platform": {
            "python_implementation": "CPython",
            "python_version": "3.5.3",
            "system": "Linux",
        },
    }
    option_manager = mock.Mock(
        registered_plugins={
            manager.PluginVersion("pycodestyle", "2.0.0", False),
            manager.PluginVersion("mccabe", "0.5.9", False),
        },
        version="3.1.0",
    )
    assert expected == debug.information(option_manager)
    pyimpl.assert_called_once_with()
    pyversion.assert_called_once_with()
    system.assert_called_once_with()


@mock.patch("flake8.main.debug.print")
@mock.patch("flake8.main.debug.information", return_value={})
@mock.patch("json.dumps", return_value="{}")
def test_print_information_no_plugins(dumps, information, print_mock):
    """Verify we print and exit only when we have plugins."""
    option_manager = mock.Mock(registered_plugins=set())
    action = debug.DebugAction(
        "--bug-report",
        dest="bug_report",
        option_manager=option_manager,
    )
    assert action(None, None, None, None) is None
    assert dumps.called is False
    assert information.called is False
    assert print_mock.called is False


@mock.patch("flake8.main.debug.print")
@mock.patch("flake8.main.debug.information", return_value={})
@mock.patch("json.dumps", return_value="{}")
def test_print_information(dumps, information, print_mock):
    """Verify we print and exit only when we have plugins."""
    plugins = [
        manager.PluginVersion("pycodestyle", "2.0.0", False),
        manager.PluginVersion("mccabe", "0.5.9", False),
    ]
    option_manager = mock.Mock(registered_plugins=set(plugins))
    action = debug.DebugAction(
        "--bug-report",
        dest="bug_report",
        option_manager=option_manager,
    )
    with pytest.raises(SystemExit):
        action(None, None, None, None)
    print_mock.assert_called_once_with("{}")
    dumps.assert_called_once_with({}, indent=2, sort_keys=True)
    information.assert_called_once_with(option_manager)



"""Test aggregation of config files and command-line options."""
import argparse
import os

import pytest

from flake8.main import options
from flake8.options import aggregator
from flake8.options import config
from flake8.options import manager

CLI_SPECIFIED_CONFIG = "tests/fixtures/config_files/cli-specified.ini"


@pytest.fixture
def optmanager():
    """Create a new OptionManager."""
    prelim_parser = argparse.ArgumentParser(add_help=False)
    options.register_preliminary_options(prelim_parser)
    option_manager = manager.OptionManager(
        prog="flake8",
        version="3.0.0",
        parents=[prelim_parser],
    )
    options.register_default_options(option_manager)
    return option_manager


def test_aggregate_options_with_config(optmanager):
    """Verify we aggregate options and config values appropriately."""
    arguments = [
        "flake8",
        "--select",
        "E11,E34,E402,W,F",
        "--exclude",
        "tests/*",
    ]
    config_finder = config.ConfigFileFinder(
        "flake8", config_file=CLI_SPECIFIED_CONFIG
    )
    options, args = aggregator.aggregate_options(
        optmanager, config_finder, arguments
    )

    assert options.select == ["E11", "E34", "E402", "W", "F"]
    assert options.ignore == ["E123", "W234", "E111"]
    assert options.exclude == [os.path.abspath("tests/*")]


def test_aggregate_options_when_isolated(optmanager):
    """Verify we aggregate options and config values appropriately."""
    arguments = [
        "flake8",
        "--select",
        "E11,E34,E402,W,F",
        "--exclude",
        "tests/*",
    ]
    config_finder = config.ConfigFileFinder("flake8", ignore_config_files=True)
    optmanager.extend_default_ignore(["E8"])
    options, args = aggregator.aggregate_options(
        optmanager, config_finder, arguments
    )

    assert options.select == ["E11", "E34", "E402", "W", "F"]
    assert sorted(options.ignore) == [
        "E121",
        "E123",
        "E126",
        "E226",
        "E24",
        "E704",
        "E8",
        "W503",
        "W504",
    ]
    assert options.exclude == [os.path.abspath("tests/*")]



"""Integration tests for plugin loading."""
from flake8.main import application

LOCAL_PLUGIN_CONFIG = "tests/fixtures/config_files/local-plugin.ini"
LOCAL_PLUGIN_PATH_CONFIG = "tests/fixtures/config_files/local-plugin-path.ini"


class ExtensionTestPlugin:
    """Extension test plugin."""

    name = "ExtensionTestPlugin"
    version = "1.0.0"

    def __init__(self, tree):
        """Construct an instance of test plugin."""

    def run(self):
        """Do nothing."""

    @classmethod
    def add_options(cls, parser):
        """Register options."""
        parser.add_option("--anopt")


class ReportTestPlugin:
    """Report test plugin."""

    name = "ReportTestPlugin"
    version = "1.0.0"

    def __init__(self, tree):
        """Construct an instance of test plugin."""

    def run(self):
        """Do nothing."""


def test_enable_local_plugin_from_config():
    """App can load a local plugin from config file."""
    app = application.Application()
    app.initialize(["flake8", "--config", LOCAL_PLUGIN_CONFIG])

    assert app.check_plugins is not None
    assert app.check_plugins["XE"].plugin is ExtensionTestPlugin
    assert app.formatting_plugins is not None
    assert app.formatting_plugins["XR"].plugin is ReportTestPlugin


def test_local_plugin_can_add_option():
    """A local plugin can add a CLI option."""
    app = application.Application()
    app.initialize(
        ["flake8", "--config", LOCAL_PLUGIN_CONFIG, "--anopt", "foo"]
    )

    assert app.options is not None
    assert app.options.anopt == "foo"


def test_enable_local_plugin_at_non_installed_path():
    """Can add a paths option in local-plugins config section for finding."""
    app = application.Application()
    app.initialize(["flake8", "--config", LOCAL_PLUGIN_PATH_CONFIG])

    assert app.check_plugins is not None
    assert app.check_plugins["XE"].plugin.name == "ExtensionTestPlugin2"



"""Tests for the Manager object for FileCheckers."""
import errno
from unittest import mock

import pytest

from flake8 import checker
from flake8.main.options import JobsArgument


def style_guide_mock():
    """Create a mock StyleGuide object."""
    return mock.MagicMock(
        **{
            "options.diff": False,
            "options.jobs": JobsArgument("4"),
        }
    )


def _parallel_checker_manager():
    """Call Manager.run() and return the number of calls to `run_serial`."""
    style_guide = style_guide_mock()
    manager = checker.Manager(style_guide, [], [])
    # multiple checkers is needed for parallel mode
    manager.checkers = [mock.Mock(), mock.Mock()]
    return manager


def test_oserrors_cause_serial_fall_back():
    """Verify that OSErrors will cause the Manager to fallback to serial."""
    err = OSError(errno.ENOSPC, "Ominous message about spaceeeeee")
    with mock.patch("_multiprocessing.SemLock", side_effect=err):
        manager = _parallel_checker_manager()
        with mock.patch.object(manager, "run_serial") as serial:
            manager.run()
    assert serial.call_count == 1


@mock.patch("flake8.checker._multiprocessing_is_fork", return_value=True)
def test_oserrors_are_reraised(is_windows):
    """Verify that unexpected OSErrors will cause the Manager to reraise."""
    err = OSError(errno.EAGAIN, "Ominous message")
    with mock.patch("_multiprocessing.SemLock", side_effect=err):
        manager = _parallel_checker_manager()
        with mock.patch.object(manager, "run_serial") as serial:
            with pytest.raises(OSError):
                manager.run()
    assert serial.call_count == 0


def test_multiprocessing_is_disabled():
    """Verify not being able to import multiprocessing forces jobs to 0."""
    style_guide = style_guide_mock()
    with mock.patch("flake8.checker.multiprocessing", None):
        manager = checker.Manager(style_guide, [], [])
        assert manager.jobs == 0


def test_make_checkers():
    """Verify that we create a list of FileChecker instances."""
    style_guide = style_guide_mock()
    files = ["file1", "file2"]
    checkplugins = mock.Mock()
    checkplugins.to_dictionary.return_value = {
        "ast_plugins": [],
        "logical_line_plugins": [],
        "physical_line_plugins": [],
    }
    with mock.patch("flake8.checker.multiprocessing", None):
        manager = checker.Manager(style_guide, files, checkplugins)

    with mock.patch("flake8.utils.filenames_from") as filenames_from:
        filenames_from.side_effect = [["file1"], ["file2"]]
        with mock.patch("flake8.utils.fnmatch", return_value=True):
            with mock.patch("flake8.processor.FileProcessor"):
                manager.make_checkers()

    assert manager._all_checkers
    for file_checker in manager._all_checkers:
        assert file_checker.filename in files
    assert not manager.checkers  # the files don't exist
